__all__ = ["is_none", "normalise_to_string_and_uppercase", "list"]

from ...constant.various import NONE_VALUE
from . import list


def is_none(thread: str | None) -> bool:
    return thread is None or thread == NONE_VALUE


def normalise_to_string_and_uppercase(thread: str | None) -> str:
    """The thread list iterator only supports strings and numbers and not None, hence the renaming to "NONE"."""

    return NONE_VALUE if is_none(thread) else str(thread).upper()



from ... import error
from ...model.thread_item import ThreadItem
from ...model.timer_base import TimerBase


def try_get_thread_item_and_index(timer: TimerBase, thread: str) -> tuple[ThreadItem, int] | tuple[None, None]:
    try:
        index_counter = 0
        for thread_item in timer.threads:
            # Check if thread already exists in list. Note that this expects a normalised thread (i.e. string and not None) and already in uppercase so it evaluates as intended.
            if thread_item.name == thread:
                # Make sure only the first match is returned, yet the main function is not designed to allow input of duplicates.
                return thread_item, index_counter
            else:
                index_counter += 1
        return None, None
    except Exception:
        error.message_for_action("in the Timer's lookup module", thread)
        return None, None


def add(timer: TimerBase, thread: str, start_time: int, decimals: int) -> None:
    try:
        timer.threads.append(ThreadItem(name=thread, start_time=start_time, decimals=decimals))
    except Exception:
        error.message_for_action("when trying to add entry to the Timer's thread list", thread)


def remove(timer: TimerBase, entry_index: int) -> None:
    try:
        timer.threads.pop(entry_index)
    except Exception:
        error.message_for_action(
            f"when trying to remove entry from the Timer's thread list for entry index \"{entry_index}\"")



from colorist import Color

from .. import error
from ..constant.decimals import DEFAULT, MAXIMUM, MINIMUM
from ..model.timer_base import TimerBase


def mediate(timer: TimerBase, decimals: int | None) -> int:
    """If the start function doesn't have decimals defined, then use the decimals value defined when the Timer() was initiated."""

    return timer.decimals if decimals is None else validate_and_normalise(decimals)


def validate_and_normalise(decimals: int | float | None) -> int:
    try:
        if not isinstance(decimals, (int, float)) or decimals is None:
            print(f"{Color.YELLOW}Timer: Decimals set to default {DEFAULT} due to invalid input.{Color.OFF}")
            return DEFAULT
        elif decimals > MAXIMUM:
            print(f"{Color.YELLOW}Timer: Decimals set to {MAXIMUM} as the Timer doesn't support more than {MAXIMUM} decimals (i.e. nanoseconds).{Color.OFF}")
            return MAXIMUM
        elif MINIMUM <= decimals <= MAXIMUM:
            if decimals % 1 == 0:
                return int(decimals)
            decimals_rounded = int(round(decimals))
            print(f"{Color.YELLOW}Timer: Decimals {decimals} set to {decimals_rounded} as a floating number is invalid input.{Color.OFF}")
            return decimals_rounded
        else:
            print(f"{Color.YELLOW}Timer: Decimals set to default {DEFAULT} due to invalid input.{Color.OFF}")
            return DEFAULT
    except Exception:
        error.message_for_action(f"when trying to verify the Timer's decimals input \"{decimals}\"")
        return DEFAULT



from .. import helper
from .elapsed_time_fractions import ElapsedTimeFractions


class TimeFractions:
    __slots__ = ["time"]

    def __init__(self, elapsed_time_ns: int) -> None:
        self.time: ElapsedTimeFractions = helper.time_fractions.calculate_time_fractions(elapsed_time_ns)

    def count_microseconds_to_float(self) -> float:
        return self.time.microseconds + self.time.nanoseconds / 1000

    def count_milliseconds_to_float(self) -> float:
        """This could potentially be faster by dividing self.microseconds by a 1000 directly, yet we don't want to lose precision in the decimals."""

        return self.time.milliseconds + self.count_microseconds_to_float() / 1000

    def count_seconds_to_float(self) -> float:
        """This could potentially be faster by dividing self.milliseconds by a 1000 directly, yet we don't want to lose precision in the decimals."""

        return self.time.seconds + self.count_milliseconds_to_float() / 1000

    def seconds_rounded(self) -> float:
        """For instance, if 2 seconds and 567 milliseconds, ensure it'll be rounded up to 3 seconds."""

        return int(round(self.count_seconds_to_float(), 0))

    def count_minutes_to_seconds(self) -> float:
        return self.time.minutes * 60 + self.count_seconds_to_float()



from ..model.elapsed_time_fractions import ElapsedTimeFractions


def calculate_time_fractions(elapsed_time_ns: int) -> ElapsedTimeFractions:
    """Elapsed time is in nanoseconds and should be calculated as difference between start and stop time using on the time.perf_counter_ns() function."""

    microseconds, nanoseconds = divmod(elapsed_time_ns, 1000)
    # As divmod() can be slow, let's return 0s as a tuple if divmod() isn't needed:
    milliseconds, microseconds = divmod(microseconds, 1000) if microseconds > 0 else (0, 0)
    seconds, milliseconds = divmod(milliseconds, 1000) if milliseconds > 0 else (0, 0)
    minutes, seconds = divmod(seconds, 60) if seconds > 0 else (0, 0)
    hours, minutes = divmod(minutes, 60) if minutes > 0 else (0, 0)
    days, hours = divmod(hours, 24) if hours > 0 else (0, 0)

    return ElapsedTimeFractions(
        nanoseconds=int(nanoseconds),
        microseconds=int(microseconds),
        milliseconds=int(milliseconds),
        seconds=int(seconds),
        minutes=int(minutes),
        hours=int(hours),
        days=int(days))



__all__: list[str] = []

from .. import error, helper
from ..model.time_fractions import TimeFractions
from ..model.timer_base import TimerBase


def start(timer: TimerBase, thread: str | None, start_time: int, decimals: int | None) -> None:
    try:
        thread = helper.thread.normalise_to_string_and_uppercase(thread)
        decimals = helper.decimals.mediate(timer, decimals)
        thread_item, _ = helper.thread.list.try_get_thread_item_and_index(timer, thread)
        if thread_item is None:  # If no match in existing threads, create new entry in the thread list.
            helper.thread.list.add(timer, thread, start_time, decimals)
        else:
            error.start_controller(thread)
    except Exception:
        error.message_for_action("in the Timer's start thread controller", thread)


def stop(timer: TimerBase, thread: str | None, stop_time: int) -> None:
    try:
        thread = helper.thread.normalise_to_string_and_uppercase(thread)
        thread_item, entry_index = helper.thread.list.try_get_thread_item_and_index(timer, thread)
        # If there's a match in existing threads, return values to the stop function and remove the entry.
        if thread_item is not None and entry_index is not None:
            helper.thread.list.remove(timer, entry_index)
            elapsed_time = stop_time - thread_item.start_time
            time_fractions = TimeFractions(elapsed_time)
            helper.output.message(thread, time_fractions, thread_item.decimals)
        else:
            error.stop_controller(timer, thread)
    except Exception:
        error.message_for_action("in the Timer's stop thread controller", thread)



__all__: list[str] = []

from colorist import Color

from .. import helper
from ..model.timer_base import TimerBase


def start_controller(thread: str | None) -> None:
    if helper.thread.is_none(thread):
        print(f"{Color.YELLOW}Timer is running. Use .stop() to stop it.{Color.OFF}")
    else:
        print(f"{Color.YELLOW}Timer for thread {thread} is running. Use .stop({thread = }) to stop it.{Color.OFF}")


def stop_controller(timer: TimerBase, thread: str | None) -> None:
    if helper.thread.is_none(thread):
        print(f"{Color.YELLOW}Timer is not running. Use .start() to start it.{Color.OFF}")
    else:
        print(f"{Color.YELLOW}Timer for thread {thread} is not running. Use .start({thread = }) to start it.{Color.OFF}")
    if len(timer.threads) > 0:
        open_threads = [thread_item.name for thread_item in timer.threads]
        print(f"Or maybe you aren't stopping the right thread? Currently open threads: {', '.join(open_threads)}")


def message_for_action(action: str, thread: str | None = None) -> None:
    print(f"{Color.YELLOW}Timer: Something went wrong {action}{'' if helper.thread.is_none(thread) else f' for thread {thread}'}.{Color.OFF}")



from colorist import Color

from .. import error, helper
from ..model.time_fractions import TimeFractions


def message(thread: str, fractions: TimeFractions, decimals: int) -> None:
    try:
        text_intro = f"Elapsed time{'' if helper.thread.is_none(thread) else f' (thread {Color.GREEN}{thread}{Color.OFF})'}:"
        time = fractions.time
        if time.days > 0:
            # Format: 1d 2h 3m 4s
            print(f"{text_intro} {time.days}d {time.hours}h {time.minutes}m {fractions.seconds_rounded()}s")
        elif time.hours > 0:
            # Format: 1h 2m 3s
            print(f"{text_intro} {time.hours}h {time.minutes}m {fractions.seconds_rounded()}s")
        elif time.minutes > 0:
            # Format: 62.34 seconds (1m 2s)
            print(f"{text_intro} {fractions.count_minutes_to_seconds():.{decimals}f} seconds ({time.minutes}m {fractions.seconds_rounded()}s)")
        elif time.seconds > 0:
            # Format: 0.123456789 seconds
            print(f"{text_intro} {fractions.count_seconds_to_float():.{decimals}f} seconds")
        elif time.milliseconds > 0:
            # Format: 123.45 milliseconds
            print(f"{text_intro} {fractions.count_milliseconds_to_float():.{decimals}f} milliseconds")
        elif time.microseconds > 0:
            # Format: 234.56 microseconds
            print(f"{text_intro} {fractions.count_microseconds_to_float():.{decimals}f} microseconds")
        else:
            # Format: 345 nanoseconds
            print(f"{text_intro} {time.nanoseconds} nanoseconds")
    except Exception:
        error.message_for_action("in the Timer's output message module", thread)



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ThreadItem:
    """Class to define a thread's name, start time, and decimals."""

    name: str
    start_time: int
    decimals: int



from abc import ABC, abstractmethod

from .thread_item import ThreadItem


class TimerBase(ABC):
    """Abstract base class of Timer object."""

    __slots__ = ["decimals", "threads"]

    def __init__(self) -> None:
        self.decimals: int
        self.threads: list[ThreadItem]

    @abstractmethod
    def start(self, thread: str, decimals: int) -> None:
        """Starts the Timer."""

        raise NotImplementedError

    @abstractmethod
    def stop(self, thread: str) -> None:
        """Stops the Timer."""

        raise NotImplementedError



from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class ElapsedTimeFractions:
    """Class to define fractions of elapsed time."""

    nanoseconds: int
    microseconds: int
    milliseconds: int
    seconds: int
    minutes: int
    hours: int
    days: int



from __future__ import annotations

import time
from types import TracebackType

from .. import constant, controller, error, helper
from .thread_item import ThreadItem
from .timer_base import TimerBase


class Timer(TimerBase):
    __slots__ = ["threads", "decimals",
                 "context_manager_threads", "context_manager_latest_thread", "context_manager_latest_decimals",
                 "__dict__"]

    _instance = None
    _lock_init = False

    def __new__(cls, thread: str | None = None, decimals: int = constant.decimals.DEFAULT) -> TimerBase:  # type: ignore
        if not cls._instance:  # Singleton: Ensure there's only a single instance of Timer running.
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, thread: str | None = None, decimals: int = constant.decimals.DEFAULT) -> None:
        if not self._lock_init:  # Ensure that initialisation of the lists only runs the first time.
            self.threads: list[ThreadItem] = []
            self.context_manager_threads: list[str] = []
            self._lock_init = True
        self.decimals: int = decimals if decimals == constant.decimals.DEFAULT else helper.decimals.validate_and_normalise(
            decimals)
        self.context_manager_latest_thread: str = helper.thread.normalise_to_string_and_uppercase(thread)
        self.context_manager_latest_decimals: int = self.decimals

    def __enter__(self) -> Timer:
        self.context_manager_threads.append(self.context_manager_latest_thread)
        self.start(self.context_manager_latest_thread, self.context_manager_latest_decimals)
        return self

    def __exit__(self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None) -> None:
        last_context_manager_thread = self.context_manager_threads.pop()
        self.stop(last_context_manager_thread)

    def start(self, thread: str | None = None, decimals: int | None = None) -> None:
        try:
            start_time = time.perf_counter_ns()  # For precision, this is the first operation of the function.
            controller.start(self, thread, start_time, decimals)
        except Exception:
            error.message_for_action("when trying to start the Timer", thread)

    def stop(self, thread: str | None = None) -> None:
        try:
            stop_time = time.perf_counter_ns()  # For precision, this is the first operation of the function.
            controller.stop(self, thread, stop_time)
        except Exception:
            error.message_for_action("when trying to stop the Timer", thread)
